# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

r"""Invenio-Records is a metadata storage module.

In a few words, a `record`_ is basically a structured collection of fields and
values (metadata) which provides information about other data.

.. _record: https://en.wikipedia.org/wiki/Record_(computer_science)

A record (and each revision) is identified by a unique `UUID`_, as most of the
others entities in Invenio.

.. _UUID: https://en.wikipedia.org/wiki/Universally_unique_identifier

Invenio-Records is a core component of Invenio and it provides a way to create,
update and delete records. Records are versioned, to keep track of
modifications and to be able to revert back to a specific revision.

When creating or updating a record, if the record contains a schema definition,
the record data will be validated against its schema. Moreover, data format can
for each field be also validated.

When deleting a record, two options are available:

* **soft deletion**: record will be deletes but keeping its identifier and
  history, to ensure that the same record's identifier cannot be reused, and
  that older revisions can be retrieved.
* **hard deletion**: record will be completely deleted with its history.

Records creation and update can be validated if the schema is provided.

Further documentation available Documentation:
https://invenio-records.readthedocs.io/

Initialization
--------------

Create a Flask application:

>>> import os
>>> db_url = os.environ.get('SQLALCHEMY_DATABASE_URI', 'sqlite://')
>>> from flask import Flask
>>> app = Flask('myapp')
>>> app.config.update({
...     'SQLALCHEMY_DATABASE_URI': db_url,
...     'SQLALCHEMY_TRACK_MODIFICATIONS': False,
... })

Initialize Invenio-Records dependencies and Invenio-Records itself:

>>> from invenio_db import InvenioDB
>>> ext_db = InvenioDB(app)
>>> from invenio_records import InvenioRecords
>>> ext_records = InvenioRecords(app)

The following examples needs to run in a Flask application context, so
let's push one:

>>> app.app_context().push()

Also, for the examples to work we need to create the database and tables (note,
in this example we use an in-memory SQLite database by default):

>>> from invenio_db import db
>>> db.create_all()

CRUD operations
---------------

Creation
~~~~~~~~

Let's **create** a very simple record:

>>> from invenio_records import Record
>>> record = Record.create({"title": "The title of the record"})
>>> db.session.commit()
>>> assert record.revision_id == 0

A new row has been added to the database, in the table ``records_metadata``:
this corresponds to the record metadata, first version (version 1).

Update
~~~~~~

Let's try to **update** the previously created record with new data. This will
create a new version of the previous with the same ``uuid`` but incremented
version/revision id.
Update the record and **commit** the changes to apply them to the record:

>>> record['title'] = 'The title of the 2nd version of the record'
>>> record = record.commit()  # validate new data and store changes
>>> db.session.commit()
>>> assert record.revision_id == 1

A second row has been added, version 2. You can access to the different
versions by doing:

>>> rec_v1 = record.revisions[0]
>>> rec_v2 = record.revisions[1]

Reverting
~~~~~~~~~

To **restore** the first version of the record, just:

>>> record = record.revert(0)
>>> db.session.commit()
>>> assert record.revision_id == 2

Patch
~~~~~

It is also possible to **patch** a record to perform multiple operations in one
shot:

>>> record = Record.create({"title": "First title"})
>>> db.session.commit()
>>> assert len(record.revisions) == 1

>>> ops = [
...     {"op": "replace", "path": "/title", "value": "Title first record"},
...     {"op": "add", "path": "/description", "value": "Record description"}
... ]

>>> record = record.patch(ops)
>>> record = record.commit()
>>> db.session.commit()
>>> assert len(record.revisions) == 2

See `JSON Patch <http://jsonpatch.com/>`_ documentation to have nice examples.

Deletion
~~~~~~~~

Let's create another record and then **soft delete** it:

>>> record = Record.create({"title": "Record to be deleted"})
>>> db.session.commit()
>>> record['title'] = 'Record to be deleted version 2'
>>> record = record.commit()
>>> db.session.commit()

>>> deleted = record.delete()

There is only one row left in the database corresponding to this record. Notice
that the ``json`` column is empty, but the ``uuid`` is still there. This
ensures uniqueness.
The record can be retrieved by doing:

>>> deleted = Record.get_record(record.id, with_deleted=True)
>>> assert deleted.id == record.id

Let's **hard delete** it, completely:

>>> deleted = record.delete(force=True)

Now, try to retrieve it, it will throw an exception.

>>> Record.get_record(record.id,
...                   with_deleted=True)  # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
  ...
NoResultFound: No row was found for one()

Record validation
-----------------

When creating or updating a record, the input data can be validated to ensure
that it is conform to a specified schema and values formats are respected.
The validation is provided by the
`jsonschema <https://python-jsonschema.readthedocs.io>`_ library.

How ``jsonschema`` works
~~~~~~~~~~~~~~~~~~~~~~~~

* **Format checker:** create a custom format checker (or use one of the
  available), for example to  validate if the first letter of a string is
  uppercase:

  >>> from jsonschema import FormatChecker
  >>> from jsonschema.validators import Draft4Validator
  >>> checker = FormatChecker()
  >>> f = checker.checks("uppercaseFirstLetter")(lambda value: value[0]
  ...                                             .isupper())
  >>> validator = Draft4Validator({"format": "uppercaseFirstLetter"},
  ...                             format_checker=checker)

  Now, let's try it out:

  >>> validator.validate("Title of the record")

  Does not throw any exception, because the data is valid, the first letter is
  uppercase.

  >>> validator.validate(
  ...               "title of the record")  # doctest: +IGNORE_EXCEPTION_DETAIL
  Traceback (most recent call last):
    ...
  ValidationError: 'title of the record' is not a 'uppercaseFirstLetter'
    ...

  This raises a ValidationError error exception, because the first letter is
  lowercase.

* **Schema validator:** create a validator to ensure that the input data
  structure, fields and types conform to a specific schema.

  >>> schema = {
  ...     'type': 'object',
  ...     'properties': {
  ...         'title': { 'type': 'string' },
  ...         'description': { 'type': 'string' }
  ...     },
  ...     'required': ['title']
  ... }

  Try to validate a record without the field `title`, which is required.

  >>> from jsonschema.validators import validate
  >>> record = {"description": "Description but no title"}
  >>> validate(record, schema)  # doctest: +IGNORE_EXCEPTION_DETAIL
  Traceback (most recent call last):
    ...
  ValidationError: 'title' is a required property
    ...

If the JSON schema is not defined inside the JSON itself, like in the example,
but it is defined somewhere else (e.g. any schema provider service), the record
should contain the ``$ref`` field with the URI link to the schema definition.
Record provides a method :py:meth:`api.RecordBase.replace_refs` that
will resolve the URI in the ``$ref`` field and return a new Record with the
schema definition injected.

Invenio-Records validation
~~~~~~~~~~~~~~~~~~~~~~~~~~

Let's put everything together and create a record with validation and format
checking: define a schema with a mandatory ``title`` field and a validation
format for the ``title`` field.

>>> from jsonschema import FormatChecker
>>> checker = FormatChecker()
>>> f = checker.checks("uppercaseFirstLetter")(lambda value: value[0]
...                                             .isupper())
>>> schema = {
...     'type':'object',
...     'properties': {
...         'title': {
...             'type':'string',
...             'format': 'uppercaseFirstLetter'
...         },
...         'description': {
...             'type':'string'
...         }
...     },
...     'required': ['title']
... }

Create a new record with an invalid value format for the ``title`` field.
Notice that the ``schema`` must be defined in the record with the field
``$schema`` and the format checker must be passed as ``kwarg`` argument with
the key ``format_checker``, to be taken into account by the `jsonschema`
library.

>>> record = {
...     "$schema": schema,
...     "title": "title of this record",  # first letter is lowercase
...     "description": "Description of this record"
... }
>>> rec = Record.create(record,
...                format_checker=checker)  # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
  ...
ValidationError: 'title of this record' is not a 'uppercaseFirstLetter'
  ...

Create a new record without the ``title`` field:

>>> record = {
...     "$schema": schema,
...     "description": "Description of this record without a title"
... }
>>> rec = Record.create(record,
...                format_checker=checker)  # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
  ...
ValidationError: 'title' is a required property
  ...

Signals
-------
Invenio-Records provides several types of signals and they can be used to react
to events to read or modify data before or after an operation.

Events are sent in case of:

* record creation, before and after
* record update, before and after
* record deletion, before and after
* record revert, before and after

Let's modify the record before creation and verify, after creation, that the
record has been correctly modified:

>>> from invenio_records.signals import (before_record_insert, \
...                                      after_record_insert)
>>> def before_record_creation_add_flag(sender, *args, **kwargs):
...     record = kwargs['record']
...     record['created_with'] = 'Invenio'
...
>>> listener = before_record_insert.connect(before_record_creation_add_flag)
>>> def after_record_creation(sender, *args, **kwargs):
...     record = kwargs['record']
...     assert 'created_with' in record
...
>>> listener = after_record_insert.connect(after_record_creation)
>>> rec_events = Record.create({"title": "My new record"})
>>> db.session.commit()

See :doc:`api` for extensive API documentation.
"""


from .api import Record
from .ext import InvenioRecords

__version__ = '1.7.2'

__all__ = (
    'InvenioRecords',
    'Record',
    '__version__',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record module signals."""

from blinker import Namespace

_signals = Namespace()

before_record_insert = _signals.signal('before-record-insert')
"""Signal is sent before a record is inserted.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.
Example event listener (subscriber) implementation:

.. code-block:: python

    def listener(sender, *args, **kwargs):
        record = kwargs['record']
        # do something with the record

    from invenio_records.signals import before_record_insert
    before_record_insert.connect(listener)
"""

after_record_insert = _signals.signal('after-record-insert')
"""Signal sent after a record is inserted.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.

.. note::

   Do not perform any modification to the record here: they will be not
   persisted.
"""

before_record_update = _signals.signal('before-record-update')
"""Signal is sent before a record is updated.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.
"""

after_record_update = _signals.signal('after-record-update')
"""Signal sent after a record is updated.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.

.. note::

   Do not perform any modification to the record here: they will be not
   persisted.
"""

before_record_delete = _signals.signal('before-record-delete')
"""Signal is sent before a record is deleted.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.
"""

after_record_delete = _signals.signal('after-record-delete')
"""Signal sent after a record is deleted.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.

.. note::

   Do not perform any modification to the record here: they will be not
   persisted.
"""

before_record_revert = _signals.signal('before-record-revert')
"""Signal is sent before a record is reverted.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.
"""

after_record_revert = _signals.signal('after-record-revert')
"""Signal sent after a record is reverted.

When implementing the event listener, the record data can be retrieved from
`kwarg['record']`.

.. note::

   Do not perform any modification to the record here: they will be not
   persisted.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2021      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Record validators."""

from jsonschema.validators import Draft4Validator, extend, validator_for

PartialDraft4Validator = extend(Draft4Validator, {'required': None})
"""Partial JSON Schema (draft 4) validator.

Special validator that contains the same validation rules of Draft4Validator,
except for required fields.
"""


def _generate_legacy_type_checks(types):
    """Generate new-style type mappings from old-style type mappings.

    Based on the function `jsonschema.validators._generate_legacy_type_checks`
    from jsonschema 3.x.

    :param types: A mapping of type names to their Python types
    :returns: A dictionary of definitions that can be passed to `TypeChecker`s
    """

    def gen_type_check(pytypes):
        def type_check(_checker, instance):
            return isinstance(instance, pytypes)

        return type_check

    return {
        typename: gen_type_check(pytypes)
        for (typename, pytypes) in types.items()
    }


def _create_validator(schema, base_validator_cls=None, custom_checks=None):
    """Create a fitting jsonschema validator class.

    :param schema: The schema for which to create a fitting validator
    :param base_validator_cls: The base :class:`jsonschema.protocols.Validator`
        class to base the new validator class on -- if not specified, the base
        class will be determined by the given schema.
    :param custom_checks: A dictionary with type names and Python types
        to check against, e.g. {"string": str, "object": dict}
    :returns: An fitting :class:`jsonschema.protocols.Validator` class
    """
    validator_cls = base_validator_cls or validator_for(schema)
    if custom_checks:
        type_checker = validator_cls.TYPE_CHECKER.redefine_many(
            _generate_legacy_type_checks(custom_checks)
        )

        validator_cls = extend(
            validator_cls,
            type_checker=type_checker,
        )

    return validator_cls



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create records branch."""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '1095cdf9f350'
down_revision = None
branch_labels = (u'invenio_records',)
depends_on = 'dbdbc1b19cf2'


def upgrade():
    """Upgrade database."""


def downgrade():
    """Downgrade database."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test for related model system fields."""

from collections import namedtuple
from copy import deepcopy

import pytest

from invenio_records.api import Record
from invenio_records.systemfields import RelatedModelField, \
    RelatedModelFieldContext, SystemFieldsMixin, relatedmodelfield


#
# Fixtures
#
@pytest.fixture()
def patch_inspect(monkeypatch):
    """Monkey patch sqlalchemy.inspect"""
    # Used to avoid having to create DB tables
    MockResult = namedtuple('MockResult', ['persistent'])

    def mock_inspect(obj):
        return MockResult(persistent=getattr(obj, 'persistent', False))

    monkeypatch.setattr(
        relatedmodelfield,
        'inspect',
        mock_inspect)

    return mock_inspect


@pytest.fixture()
def patch_db(monkeypatch):
    """Monkey patch session_merge"""
    # Used to avoid having to create DB tables
    class MockSession:
        @staticmethod
        def merge(obj):
            obj = deepcopy(obj)
            obj.persistent = True
            return obj

    class MockDb:
        session = MockSession

    monkeypatch.setattr(
        relatedmodelfield,
        'db',
        MockDb
    )


@pytest.fixture()
def MyModel():
    """Create a model class."""
    class MyModel:
        def __init__(self, val):
            self._val = val

        @classmethod
        def dump_obj(cls, field, record, obj):
            field.set_dictkey(record, {'akey': obj._val})

        @classmethod
        def load_obj(cls, field, record):
            data = field.get_dictkey(record)
            if data:
                return cls(data['akey'])
            return None

        def __eq__(self, o):
            return self._val == o._val

    return MyModel


@pytest.fixture()
def MyRecord(MyModel):
    """A record with a related model system field."""
    class MyRecord(Record, SystemFieldsMixin):
        myobj = RelatedModelField(MyModel)

    return MyRecord


#
# Tests
#
def test_field(MyRecord, MyModel):
    """Test class behaviour."""
    assert isinstance(MyRecord.myobj, RelatedModelFieldContext)


def test_setting_related_obj(testapp, db, MyRecord, MyModel):
    """Set the related object"""
    # via attr access
    record = MyRecord({})
    record.myobj = MyModel('a')
    assert record['myobj'] == {'akey': 'a'}
    record.myobj = MyModel('b')
    assert record['myobj'] == {'akey': 'b'}

    # via __init__() with a myobj argument
    record = MyRecord({}, myobj=MyModel('c'))
    assert record['myobj'] == {'akey': 'c'}

    # via create() with a myobj argument
    record = MyRecord.create({}, myobj=MyModel('c'))
    assert record['myobj'] == {'akey': 'c'}


def test_change_key(testapp, db, MyModel):
    """Change key where the record is stored."""
    class Record1(Record, SystemFieldsMixin):
        myobj = RelatedModelField(MyModel, key='custom')

    assert Record1({}, myobj=MyModel('a'))['custom'] == {'akey': 'a'}


def test_custom_dump_load(testapp, db, MyModel):
    """Change custom dump/load function."""
    # A custom dumper function
    def my_dump(field, record, obj):
        field.set_dictkey(record, {'custom': obj._val})

    # A custom loader function
    def my_load(field, record):
        data = field.get_dictkey(record)
        if data:
            return MyModel(data['custom'])
        return None

    class Record1(Record, SystemFieldsMixin):
        myobj = RelatedModelField(MyModel, load=my_load, dump=my_dump)

    # Dumping
    record = Record1({}, myobj=MyModel('a'))
    assert record['myobj'] == {'custom': 'a'}

    # Loading
    record = Record1(dict(record))
    assert record.myobj == MyModel('a')


def test_serialize_on_commit(testapp, db, MyRecord, MyModel):
    """Test serialize on commit."""
    obj = MyModel('a')
    record = MyRecord.create({}, myobj=obj)
    assert record['myobj'] == {'akey': 'a'}

    # Change related object outside the APIs
    obj._val = 'b'
    assert record['myobj'] == {'akey': 'a'}

    # Now commit and check that latest value are serialized.
    record.commit()
    assert record['myobj'] == {'akey': 'b'}


def test_required(testapp, db, MyModel, MyRecord):
    """Test required property."""
    class Record1(Record, SystemFieldsMixin):
        myobj = RelatedModelField(MyModel, required=True)
    # Test that an error is raised when required
    pytest.raises(RuntimeError, Record1.create({}).commit)
    # Test that no error is raised (MyRecord does not require the obj)
    assert MyRecord.create({}).commit() == {}


def test_field_context(MyModel):
    """Test the field context."""
    class MyFieldCtx(RelatedModelFieldContext):
        def test(self):
            return True

    class Record1(Record, SystemFieldsMixin):
        myobj = RelatedModelField(MyModel, context_cls=MyFieldCtx)

    assert Record1.myobj.test()


def test_ctx_session_merge(MyRecord, MyModel, patch_inspect, patch_db):
    """Test the session_merge."""
    record = MyRecord(dict(MyRecord({}, myobj=MyModel('a'))))
    assert patch_inspect(record.myobj).persistent is False
    MyRecord.myobj.session_merge(record)
    assert patch_inspect(record.myobj).persistent is True



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
# Copyright (C) 2021 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test for system fields."""

import pytest

from invenio_records.api import Record
from invenio_records.dumpers import ElasticsearchDumper
from invenio_records.systemfields import ConstantField, DictField, \
    SystemField, SystemFieldsMixin


#
# Fixtures
#
@pytest.fixture()
def testschema():
    """The JSONSchema used for tests."""
    return {
        "type": "object",
        "properties": {
            "title": {"type": "string"},
        },
        "required": ["title"]
    }


@pytest.fixture()
def SystemRecord():
    """System record is a base class."""
    class SystemRecord(Record, SystemFieldsMixin):
        # Only defined in SystemRecord:
        test = ConstantField("test", "test")
        # Defined in both SystemRecord and MyRecord:
        base = ConstantField("base", "systemrecord")
    return SystemRecord


@pytest.fixture()
def MyRecord(SystemRecord, testschema):
    """My record class inheriting from SystemRecord."""
    class MyRecord(SystemRecord):
        # Defined in only MyRecord
        schema = ConstantField("$schema", testschema)
        # Defined in both SystemRecord and MyRecord
        base = ConstantField("base", "myrecord")
    return MyRecord


@pytest.fixture()
def record(MyRecord, db):
    """Create a record instance of MyRecord."""
    return MyRecord.create({'title': 'test'})


@pytest.fixture()
def sysrecord(SystemRecord, db):
    """Create a record instance of SystemRecord."""
    return SystemRecord.create({'title': 'test'})


@pytest.fixture()
def ExtensionRecord():
    """Create an ExtensionRecord class to test extensions being called."""
    class ExtensionField(SystemField):
        called = []

        def pre_init(self, record, data, model=None, **kwargs):
            self.called.append('pre_init')

        def post_init(self, record, data, model=None, field_data=None):
            self.called.append('post_init')

        def pre_dump(self, record, dumper=None):
            self.called.append('pre_dump')

        def post_load(self, record, loader=None):
            self.called.append('post_load')

        def pre_create(self, record):
            self.called.append('pre_create')

        def post_create(self, record):
            self.called.append('post_create')

        def pre_commit(self, record):
            self.called.append('pre_commit')

        def post_commit(self, record):
            self.called.append('post_commit')

        def pre_delete(self, record, force=False):
            self.called.append('pre_delete')

        def post_delete(self, record, force=False):
            self.called.append('post_delete')

        def pre_revert(self, record, revision):
            self.called.append('pre_revert')

        def post_revert(self, new_record, revision):
            self.called.append('post_revert')

    class ExtensionRecord(Record, SystemFieldsMixin):
        dumper = ElasticsearchDumper()
        ext = ExtensionField()

    return ExtensionRecord


#
# Tests
#
def test_constant_field(testapp, record, MyRecord, testschema):
    """Test the constant field."""
    # Test that constant field value was set, and is accessible
    assert record['$schema'] == testschema
    assert record.schema == testschema

    # Test that constant fields cannot have values assigned.
    try:
        record.schema = testschema
        assert False
    except AttributeError:
        pass

    # Explicit remove the set value ($schema) after the record creation.
    record = MyRecord({'title': 'test'})
    del record['$schema']

    # Now test, that accessing the field returns None.
    assert record.schema is None


def test_constant_field_deletion(testapp, db, record, MyRecord, testschema):
    """Test the constant field."""
    # Test that constant field value was set, and is accessible
    db.session.commit()
    record.delete()
    db.session.commit()

    record = MyRecord.get_record(record.id, with_deleted=True)
    assert record.schema is None


def test_field_overwriting(testapp, record, sysrecord):
    """Test that field overwriting."""
    # field 'base' is defined in both classes, thus MyRecord overwrites
    # SystemRecord.
    assert record.base == 'myrecord'
    assert sysrecord.base == 'systemrecord'


def test_field_inheritance(testapp, record, sysrecord):
    """Test that field inheritance."""
    # field 'test' is defined only in base class, but available on both.
    assert record.test == 'test'
    assert sysrecord.test == 'test'


def test_field_class_access(testapp, MyRecord):
    """Test that class access is returning the field."""
    assert isinstance(MyRecord.schema, ConstantField)
    assert isinstance(MyRecord.test, ConstantField)
    assert isinstance(MyRecord.base, ConstantField)


def test_attrname_injection(testapp, MyRecord):
    """Test that class access is returning the field."""
    assert MyRecord.schema.attr_name == 'schema'
    assert MyRecord.test.attr_name == 'test'
    assert MyRecord.base.attr_name == 'base'


def test_systemfields_mro(testapp):
    """Test overwriting of system fields according to MRO."""
    class A(Record, SystemFieldsMixin):
        test = ConstantField("test", "a")

    class B(A):
        test = ConstantField("test", "b")

    class C(A):
        pass

    class D(B, C):
        test = ConstantField("test", "d")

    class E(B, C):
        pass

    class F(C):
        pass

    class G(C, B):
        pass

    # A defines field itself
    assert A({}).test == "a"
    # B inherits from A  (overwrites field)
    assert B({}).test == "b"
    # C inherits from A
    assert C({}).test == "a"
    # D inherits from B and C (overwrites field)
    assert D({}).test == "d"
    # E inherits from B and C
    assert E({}).test == "b"
    # F inherits from C which inherits from A.
    assert F({}).test == "a"
    # G inherits from C (which inherits from A) and B.
    assert G({}).test == "b"


def test_extension_pre_init(testapp, db, ExtensionRecord):
    """Test pre init hook."""
    rec = ExtensionRecord({})
    assert ExtensionRecord.ext.called == ['pre_init', 'post_init']


def test_extension_post_create(testapp, db, ExtensionRecord):
    """Test post create hook."""
    rec = ExtensionRecord.create({})
    assert ExtensionRecord.ext.called == [
        'pre_init', 'post_init', 'pre_create', 'post_create']


def test_extension_pre_dump(testapp, db, ExtensionRecord):
    """Test pre dump hook."""
    rec = ExtensionRecord({}).dumps()
    assert ExtensionRecord.ext.called == [
        'pre_init', 'post_init', 'pre_dump']


def test_extension_post_load(testapp, db, ExtensionRecord):
    """Test post load hook."""
    dump = ExtensionRecord({}).dumps()
    rec = ExtensionRecord.loads(dump)
    assert ExtensionRecord.ext.called == [
        'pre_init', 'post_init', 'pre_dump', 'pre_init', 'post_init',
        'post_load']


def test_extension_commit(testapp, db, ExtensionRecord):
    """Test commit hooks."""
    rec = ExtensionRecord.create({})
    rec.commit()
    assert ExtensionRecord.ext.called == [
        'pre_init', 'post_init', 'pre_create', 'post_create', 'pre_commit',
        'post_commit']


def test_extension_delete(testapp, db, ExtensionRecord):
    """Test pre/post delete hook."""
    rec = ExtensionRecord.create({})
    db.session.commit()
    rec.delete()
    assert ExtensionRecord.ext.called == [
        'pre_init', 'post_init', 'pre_create', 'post_create', 'pre_delete',
        'post_delete',
    ]


def test_extension_revert(testapp, database, ExtensionRecord):
    """Test pre/post revert hook."""
    db = database
    rec = ExtensionRecord.create({})
    db.session.commit()
    rec.commit()
    db.session.commit()
    assert rec.revision_id == 1
    rec.revert(0)
    assert ExtensionRecord.ext.called == [
         'pre_init', 'post_init', 'pre_create', 'post_create', 'pre_commit',
         'post_commit', 'pre_revert', 'pre_init', 'post_init', 'post_revert'
    ]


def test_base_systemfield_base(testapp):
    """Test default implementation of system field."""
    class TestRecord(Record, SystemFieldsMixin):
        field = SystemField()

    # Test to please the test coverage gods
    assert pytest.raises(AttributeError, getattr, TestRecord({}), 'field')
    # A core system field doesn't support assignment, so you can't pass args
    # in constructor as well
    assert pytest.raises(AttributeError, TestRecord, {}, field={})


def test_systemfield_initialization(testapp):
    """Test default implementation of system field."""
    class TestField(SystemField):
        def __set__(self, instance, value):
            instance['arg_value'] = value

    class TestRecord(Record, SystemFieldsMixin):
        afield = TestField()

    # Init method
    assert TestRecord({}, afield='testval')['arg_value'] == 'testval'

    # Create method
    record = TestRecord.create({}, afield='testval')
    assert record['arg_value'] == 'testval'


def test_default_key(testapp):
    """Test default key name."""
    class TestRecord(Record, SystemFieldsMixin):
        afield = ConstantField(value='test')

    assert TestRecord({})['afield'] == 'test'


#
# DictField
#
def test_dict_field_simple():
    """Simple tests for the DictField."""
    class Record1(Record, SystemFieldsMixin):
        metadata = DictField('metadata')

    assert isinstance(Record1.metadata, DictField)

    # Creation
    assert Record1({}, metadata={'title': 1}) == {'metadata': {'title': 1}}
    # Access key exists
    assert Record1({'metadata': {'title': 1}}).metadata == {'title': 1}
    # Access key doesn't exists
    assert Record1({}).metadata is None

    # Double tap - metadata wins over default data.
    assert Record1(
        {'metadata': {'title': 2}, 'a': 'b'},
        metadata={'title': 1}) == {'metadata': {'title': 1}, 'a': 'b'}


def test_dictfield_default_key(testapp):
    """Test default key name."""
    class TestRecord(Record, SystemFieldsMixin):
        afield = DictField()

    assert TestRecord({}, afield={'test': 1})['afield'] == {'test': 1}


def test_dict_field_complex_key():
    """More complex tests for the DictField."""
    class Record1(Record, SystemFieldsMixin):
        metadata = DictField('metadata.a.b')

    # Creation
    record = Record1({}, metadata={'title': 1})
    assert record == {'metadata': {'a': {'b': {'title': 1}}}}
    # Access key exists
    assert record.metadata == {'title': 1}
    # Access key doesn't exists
    assert Record1({}).metadata is None

    # Cannot handle lists
    pytest.raises(KeyError, Record1, {'metadata': []}, metadata={'title': 1})
    pytest.raises(
        KeyError, Record1, {'metadata': {'a': []}}, metadata={'title': 1})


def test_dict_field_create_if_missing():
    """More complex tests for the DictField."""
    class Record1(Record, SystemFieldsMixin):
        metadata = DictField('metadata.a.b', create_if_missing=False)

    # If subkey doesn't exists, it fails with create_if_missing=False
    pytest.raises(KeyError, Record1, {}, metadata={'title': 1})
    pytest.raises(KeyError, Record1, {'metadata': []}, metadata={'title': 1})
    # Subkey exists, all good:
    assert Record1({'metadata': {'a': {}}}, metadata={'title': 1})


def test_dict_field_clear_none():
    """Test dict field with clearing none/empty values."""
    class Record1(Record, SystemFieldsMixin):
        metadata = DictField('metadata', clear_none=True)

    record = Record1(
        {'a': None},
        metadata={'a': None, 'title': 1, 'b': {'c': None}}
    )
    assert record == {'a': None, 'metadata': {'title': 1}}

    record = Record1({}, metadata={'a': None})
    assert record == {'metadata': {}}


def test_dict_field_deleted(testapp, database):
    """Test dict field with clearing none/empty values."""
    db = database

    class Record1(Record, SystemFieldsMixin):
        metadata = DictField()

    record = Record1.create({}, metadata={'title': 1})
    db.session.commit()
    record.delete()
    db.session.commit()

    # Loading a deleted record
    record = Record1.get_record(record.id, with_deleted=True)
    assert record.metadata is None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests Invenio-Records JSONSchema ref resolver."""

import pytest

from invenio_records.api import Record
from invenio_records.resolver import urljoin_with_custom_scheme


def local_ref_resolver_store_factory():
    """Build local JSONSchema ref resolver store."""
    return {
        "local://authors.json": {
            "$id": "local://authors.json",
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "local://books.json": {
            "$id": "local://books.json",
            "type": "object",
            "definitions": {
                "price": {
                    "type": "string",
                },
            },
            "properties": {
                "title": {"type": "string"},
                "price": {"$ref": "#/definitions/price"},
                "authors": {"$ref": "local://authors.json"},
            }
        }
    }


@pytest.fixture(scope='module')
def app_config(app_config):
    app_config['RECORDS_REFRESOLVER_CLS'] = \
        "invenio_records.resolver.InvenioRefResolver"
    app_config['RECORDS_REFRESOLVER_STORE'] = \
        local_ref_resolver_store_factory()
    return app_config


def test_invenio_refresolver_with_local_store(db):
    """Test InvenioRefResolver with local store and complex JSONSchema."""
    data = {
        '$schema': 'local://books.json#',
        'title': 'The Lord of the Rings',
        'price': '20',
        'authors': [
            'J. R. R. Tolkien',
        ]
    }
    Record.create(data).commit()
    db.session.commit()


@pytest.mark.parametrize(
    "resolution_scope, scope, expected_output",
    [
        # Custom scheme simple cases
        ("local://books.json", "local://books.json#", "local://books.json"),
        ("local://books.json#", "local://books.json", "local://books.json"),
        ("local://books.json#", "#", "local://books.json"),
        # Custom scheme with internal refs
        (
            "local://books.json",
            "local://books.json#/definitions/price",
            "local://books.json#/definitions/price"
        ),
        (
            "local://books.json",
            "#/definitions/price",
            "local://books.json#/definitions/price"
        ),
    ]
)
def test_urljoin_with_custom_scheme(resolution_scope, scope, expected_output):
    """Test urljoin supporting custom schemas."""
    assert urljoin_with_custom_scheme(
        resolution_scope, scope) == expected_output



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations system field."""

from ...dictutils import dict_lookup, dict_set
from .errors import InvalidCheckValue, InvalidRelationValue


class RelationResult:
    """Relation access result."""

    def __init__(self, field, record):
        """Initialize the relation result."""
        self.field = field
        self.record = record

    def _lookup_id(self):
        return dict_lookup(self.record, self.value_key)

    def _lookup_data(self):
        return dict_lookup(self.record, self.key)

    def __call__(self, force=True):
        """Resolve the relation."""
        try:
            val = self._lookup_id()
            obj = self.resolve(val)
            return obj
        except KeyError:
            return None

    def __getattr__(self, name):
        """Proxy attribute access to field."""
        return getattr(self.field, name)

    def _value_check(self, value_to_check, object):
        """Checks if the value is present in the object."""
        for key, value in value_to_check.items():
            if key not in object:
                raise InvalidCheckValue(f'Invalid key {key}.')
            if isinstance(value, dict):
                self._value_check(value, object[key])
            else:
                if not isinstance(value, list):
                    raise InvalidCheckValue(
                        f"Invalid value_check value: {value}; it must be "
                        "a list"
                    )
                elif isinstance(object[key], list):
                    value_exist = set(object[key]).intersection(set(value))
                    if not value_exist:
                        raise InvalidCheckValue(
                            f"Failed cross checking value_check value "
                            f"{value} with record value {object[key]}."
                        )
                else:
                    if object[key] not in value:
                        raise InvalidCheckValue(
                            f"Failed cross checking value_check value "
                            f"{value} with record value {object[key]}."
                        )

    def validate(self):
        """Validate the field."""
        try:
            val = self._lookup_id()
            if not self.exists(val):
                raise InvalidRelationValue(f'Invalid value {val}.')
            if self.value_check:
                data = self._lookup_data()
                obj = self.resolve(data[self.field._value_key_suffix])
                self._value_check(self.value_check, obj)
        except KeyError:
            return None

    def dereference(self, keys=None, attrs=None):
        """Dereference the relation field object inside the record."""
        try:
            data = self._lookup_data()
            return self._dereference_one(
                data, keys or self.keys, attrs or self.attrs)
        except KeyError:
            return None

    def clean(self, keys=None, attrs=None):
        """Clean the dereferenced attributes inside the record."""
        try:
            data = self._lookup_data()
            return self._clean_one(
                data, keys or self.keys, attrs or self.attrs)
        except KeyError:
            return None

    def _dereference_one(self, data, keys, attrs):
        """Dereference a single object into a dict."""
        # Don't dereference if already referenced.
        if '@v' in data:
            return data

        # Get related record
        obj = self.resolve(data[self.field._value_key_suffix])
        # Inject selected key/values from related record into
        # the current record.

        # From record dictionary
        if keys is None:
            data.update({
                k: v for k, v in obj.items()
            })
        else:
            new_obj = {}
            for k in keys:
                try:
                    val = dict_lookup(obj, k)
                    if val:
                        dict_set(new_obj, k, val)
                except KeyError:
                    pass
            data.update(new_obj)

        # From record attributes (i.e. system fields)
        for a in attrs:
            data[a] = getattr(obj, a)

        # Add a version counter "@v" used for optimistic
        # concurrency control. It allows to search for all
        # outdated records and reindex them.
        data['@v'] = f'{obj.id}::{obj.revision_id}'
        return data

    def _clean_one(self, data, keys, attrs):
        """Remove all but "id" key for a dereferenced related object."""
        relation_id = data[self.field._value_key_suffix]
        data.clear()
        data[self.field._value_key_suffix] = relation_id
        return data


class RelationListResult(RelationResult):
    """Relation access result."""

    def __call__(self, force=True):
        """Resolve the relation."""
        try:
            values = self._lookup_data()
            return (self.resolve(v[self._value_key_suffix]) for v in values)
        except KeyError:
            return None

    def _lookup_id(self, data):
        return dict_lookup(data, self.field._value_key_suffix)

    def _lookup_data(self):
        data = dict_lookup(self.record, self.key)
        if self.relation_field:
            filtered_data = [
                e.get(self.relation_field) for e in data
                if self.relation_field in e
            ]
            return filtered_data
        return data

    def validate(self):
        """Validate the field."""
        try:
            values = self._lookup_data()
            if values and not isinstance(values, list):
                raise InvalidRelationValue(
                    f'Invalid value {values}, should be list.')

            for v in values:
                relation_id = self._lookup_id(v)
                if not self.exists(relation_id):
                    raise InvalidRelationValue(f'Invalid value {relation_id}.')
                if self.value_check:
                    obj = self.resolve(v[self.field._value_key_suffix])
                    self._value_check(self.value_check, obj)
        except KeyError:
            return None

    def _apply_items(self, func, keys=None, attrs=None):
        """Iterate over the list of objects."""
        # The attributes we want to get from the related record.
        keys = keys or self.keys
        attrs = attrs or self.attrs
        try:
            # Get the list of objects we have to dereference/clean.
            values = self._lookup_data()
            if values:
                for v in values:
                    # Only dereference/clean if "id" key is present in parent.
                    # Note, we control via the JSONSchema if a record can have
                    # only related records (by requiring precense of "id" key),
                    # or if you can mix non-linked records.
                    if self.field._value_key_suffix in v:
                        func(v, keys, attrs)
                return values
        except KeyError:
            return None

    def dereference(self, keys=None, attrs=None):
        """Dereference the relation field object inside the record.

        Dereferences a list of ids::

            [{"id": "eng"}]

        Into

            [{"id": "eng", "title": ..., "@v": ...}]
        """
        return self._apply_items(self._dereference_one, keys, attrs)

    def clean(self, keys=None, attrs=None):
        """Clean the dereferenced attributes inside the record.

        Reverses changes made by dereference.
        """
        return self._apply_items(self._clean_one, keys, attrs)

    def append(self, value):
        """Append a relation to the list."""
        raise NotImplementedError()

    def insert(self, index, value):
        """Insert a relation to the list."""
        raise NotImplementedError()


class RelationNestedListResult(RelationListResult):
    """Relation access result."""

    def __call__(self, force=True):
        """Resolve the relation."""
        try:
            values = self._lookup_data()
            parent_iter = []
            for inner_values in values:
                inner_iter = []
                for v in inner_values:
                    try:
                        inner_iter.append(
                            self.resolve(v[self._value_key_suffix])
                        )
                    except KeyError:
                        continue
                parent_iter.append(iter(inner_iter))
            return iter(parent_iter)
        except KeyError:
            return None

    def validate(self):
        """Validate the field."""
        try:
            values = self._lookup_data()
            if values and not isinstance(values, list):
                raise InvalidRelationValue(
                    f'Invalid value {values}, should be list.')

            for outter_v in values:
                if outter_v and not isinstance(outter_v, list):
                    raise InvalidRelationValue(
                        f'Invalid inner value {outter_v}, should be list.'
                    )
                for v in outter_v:
                    relation_id = self._lookup_id(v)
                    if not self.exists(relation_id):
                        raise InvalidRelationValue(
                            f'Invalid value {relation_id}.')
                    if self.value_check:
                        obj = self.resolve(v[self.field._value_key_suffix])
                        self._value_check(self.value_check, obj)
        except KeyError:
            return None

    def _apply_items(self, func, keys=None, attrs=None):
        """Iterate over the list of objects."""
        # The attributes we want to get from the related record.
        attrs = attrs or self.attrs
        keys = keys or self.keys
        try:
            # Get the list of objects we have to dereference/clean.
            values = self._lookup_data()
            if values:
                for outter_v in values:
                    for v in outter_v:
                        if self.field._value_key_suffix in v:
                            func(v, keys, attrs)
                return values
        except KeyError:
            return None
